import os
from cffi import FFI


ffi = FFI()
with open(os.path.join(os.path.dirname(__file__), "fdb_c.h")) as f:
    ffi.set_source("found._fdb", f.read(), libraries=["fdb_c"])


# TODO: replace fdb_c2.h with what it is, what it's useful for
with open(os.path.join(os.path.dirname(__file__), "fdb_c2.h")) as f:
    ffi.cdef(f.read())


def build(*args, **kwargs):
    ffi.compile(verbose=True)


if __name__ == "__main__":
    build()



#
# __init__.py
#
# This source file is part of the FoundationDB open source project
#
# Copyright 2013-2018 Apple Inc. and the FoundationDB project authors
# Copyright 2018-2021 Amirouche Boubekki <amirouche@hyper.dev>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from found._fdb import lib
import fdb


__VERSION__ = (0, 10, 0)

MAX_SIZE_TRANSACTION = 10 ** 7
MAX_SIZE_KEY = 10 ** 4
MAX_SIZE_VALUE = 10 ** 5


HEADER_VERSION = VERSION = 630


code = lib.fdb_select_api_version_impl(VERSION, HEADER_VERSION)
if code == 2203:
    max_supported_ver = lib.fdb_get_max_api_version()
    if HEADER_VERSION > max_supported_ver:
        msg = "This version of the FoundationDB Python binding is not supported by "
        msg += "the installed FoundationDB C library. The binding requires a library "
        msg += "that supports API version %d, but the installed library supports a "
        msg += "maximum version of %d."
        msg = msg % (HEADER_VERSION, max_supported_ver)
        raise RuntimeError(msg)
    else:
        msg = "API version %d is not supported by the installed FoundationDB C library."
        msg = msg % VERSION
        raise RuntimeError(msg)
elif code != 0:
    raise RuntimeError("FoundationDB API error ({})".format(code))


# Required to use fdb.tuple.pack and fdb.tuple.unpack
fdb._version = VERSION

from found.base import BaseFoundException  # noqa
from found.base import FoundException  # noqa
from found.base import next_prefix  # noqa
from found.base import STREAMING_MODE_WANT_ALL  # noqa
from found.base import STREAMING_MODE_ITERATOR  # noqa
from found.base import STREAMING_MODE_EXACT  # noqa
from found.base import STREAMING_MODE_SMALL  # noqa
from found.base import STREAMING_MODE_MEDIUM  # noqa
from found.base import STREAMING_MODE_LARGE  # noqa
from found.base import STREAMING_MODE_SERIAL  # noqa
from found.base import read_version  # noqa
from found.base import get  # noqa
from found.base import lt  # noqa
from found.base import lte  # noqa
from found.base import gt  # noqa
from found.base import gte  # noqa
from found.base import query  # noqa
from found.base import estimated_size_bytes  # noqa
from found.base import set_read_version  # noqa
from found.base import set  # noqa
from found.base import clear  # noqa
from found.base import add  # noqa
from found.base import bit_and  # noqa
from found.base import bit_or  # noqa
from found.base import bit_xor  # noqa
from found.base import max  # noqa
from found.base import byte_max  # noqa
from found.base import min  # noqa
from found.base import byte_min  # noqa
from found.base import set_versionstamped_key  # noqa
from found.base import set_versionstamped_value  # noqa
from found.base import transactional  # noqa
from found.base import open  # noqa

from fdb.tuple import pack  # noqa
from fdb.tuple import pack_with_versionstamp  # noqa
from fdb.tuple import unpack  # noqa
from fdb.tuple import has_incomplete_versionstamp  # noqa
from fdb.tuple import Versionstamp  # noqa

# TODO: from fdb.subspace_impl import Subspace  # noqa


def co(func):
    async def wrapper(*args, **kwargs):
        return func(*args, **kwargs)
    return wrapper



#
# Copyright (C) 2015-2022  Amirouche Boubekki <amirouche@hyper.dev>
#
# https://git.sr.ht/~amirouche/asyncio-foundationdb
#
import itertools
from math import factorial
from collections import namedtuple

import found
from found.base import FoundException

from immutables import Map


def pk(*args):
    print(*args)
    return args[-1]


# Compute the minimal set of indices required to bind any n-pattern in
# one hop.
#
# Based on https://stackoverflow.com/a/55148433/140837


def bc(n, k):
    """Binomial coefficient"""
    return factorial(n) // factorial(k) // factorial(n - k) if k < n else 0


def stringify(iterable):
    return "".join(str(x) for x in iterable)


def combinations(tab):
    out = []
    for i in range(1, len(tab) + 1):
        out.extend(stringify(x) for x in itertools.combinations(tab, i))
    assert len(out) == 2 ** len(tab) - 1
    return out


def ok(solutions, tab):
    """Check that SOLUTIONS of TAB is a correct solution"""
    cx = combinations(tab)

    for combination in cx:
        pcx = ["".join(x) for x in itertools.permutations(combination)]
        # check for existing solution
        for solution in solutions:
            if any(solution.startswith(p) for p in pcx):
                # yeah, there is an existing solution
                break
        else:
            raise Exception("failed with combination={}".format(combination))
    else:
        return True
    return False


def _compute_indices(n):
    tab = list(range(n))
    cx = list(itertools.combinations(tab, n // 2))
    for c in cx:
        L = [(i, i in c) for i in tab]
        A = []
        B = []
        while True:
            for i in range(len(L) - 1):
                if (not L[i][1]) and L[i + 1][1]:
                    A.append(L[i + 1][0])
                    B.append(L[i][0])
                    L.remove((L[i + 1][0], True))
                    L.remove((L[i][0], False))
                    break
            else:
                break
        l = [i for (i, _) in L]  # noqa
        yield tuple(A + l + B)


# helpers


def take(iterator, count):
    for _ in range(count):
        out = next(iterator)
        yield out


def drop(iterator, count):
    for _ in range(count):
        next(iterator)
    yield from iterator


class NStoreException(FoundException):
    pass


Variable = var = namedtuple('Variable', ('name',))


def is_permutation_prefix(combination, index):
    index = stringify(index)
    out = any(
        index.startswith(stringify(x)) for x in itertools.permutations(combination)
    )
    return out


_NStore = namedtuple('NStore', ('name', 'prefix', 'n', 'indices'))


def make(name, prefix, n):
    return _NStore(name, tuple(prefix), n, list(_compute_indices(n)))


def add(tx, nstore, *items, value=b''):
    assert len(items) == nstore.n, "invalid item count"
    for subspace, index in enumerate(nstore.indices):
        permutation = list(items[i] for i in index)
        key = tuple(nstore.prefix) + (subspace,) + tuple(permutation)
        found.set(tx, found.pack(key), value)


def remove(tx, nstore, *items):
    assert len(items) == nstore.n, "invalid item count"
    for subspace, index in enumerate(nstore.indices):
        permutation = list(items[i] for i in index)
        key = nstore.prefix + (subspace,) + tuple(permutation)
        found.clear(tx, found.pack(tuple(key)))


async def get(tx, nstore, *items):
    assert len(items) == nstore.n, "invalid item count"
    subspace = 0
    key = nstore.prefix + (subspace,) + tuple(items)
    out = await found.get(tx, found.pack(tuple(key)))
    out = None if out is None else out
    return out


async def select(tx, nstore, *pattern, seed=Map()):  # seed is immutable
    """Yields bindings that match PATTERN"""
    assert len(pattern) == nstore.n, "invalid item count"
    variable = tuple(isinstance(x, Variable) for x in pattern)
    # find the first index suitable for the query
    combination = tuple(x for x in range(nstore.n) if not variable[x])
    for subspace, index in enumerate(nstore.indices):
        if is_permutation_prefix(combination, index):
            break
    else:
        raise NStoreException("Oops!")
    # `index` variable holds the permutation suitable for the
    # query. `subspace` is the "prefix" of that index.
    prefix = list(pattern[i] for i in index if not isinstance(pattern[i], Variable))
    prefix = list(nstore.prefix) + [subspace] + prefix
    start = found.pack(tuple(prefix))
    end = found.next_prefix(start)
    async for key, _ in found.query(tx, start, end):
        items = found.unpack(key)[len(nstore.prefix) + 1:]
        # re-order the items
        items = tuple(items[index.index(i)] for i in range(nstore.n))
        bindings = seed
        for i, item in enumerate(pattern):
            if isinstance(item, Variable):
                bindings = bindings.set(item.name, items[i])
        yield bindings


async def where(tx, nstore, iterator, *pattern):
    assert len(pattern) == nstore.n, "invalid item count"

    async for bindings in iterator:
        # bind PATTERN against BINDINGS
        bound = []
        for item in pattern:
            # if ITEM is variable try to bind
            if isinstance(item, Variable):
                try:
                    value = bindings[item.name]
                except KeyError:
                    # no bindings
                    bound.append(item)
                else:
                    # pick the value in bindings
                    bound.append(value)
            else:
                # otherwise keep item as is
                bound.append(item)
        # hey!
        out = select(tx, nstore, *bound, seed=bindings)
        async for binding in out:
            yield binding


def query(tx, nstore, pattern, *patterns):
    out = select(tx, nstore, *pattern)
    for pattern in patterns:
        out = where(tx, nstore, out, *pattern)
    return out



#
# This source file is part of the asyncio-foundationdb open source project
#
# Copyright 2021 Amirouche Boubekki <amirouche@hyper.dev>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from aiostream import pipe, stream
import asyncio


async def pool_for_each_par_map(loop, pool, f, p, iterator):
    zx = stream.iterate(iterator)
    zx = zx | pipe.map(lambda x: loop.run_in_executor(pool, p, x))
    async with zx.stream() as streamer:
        limit = pool._max_workers
        unfinished = []
        while True:
            tasks = []
            for i in range(limit):
                try:
                    task = await streamer.__anext__()
                except StopAsyncIteration:
                    limit = 0
                else:
                    tasks.append(task)
            tasks = tasks + list(unfinished)
            assert len(tasks) <= pool._max_workers
            if not tasks:
                break
            finished, unfinished = await asyncio.wait(
                tasks, return_when=asyncio.FIRST_COMPLETED
            )
            for finish in finished:
                out = finish.result()
                f(out)
            limit = len(finished)



#
# found/bstore.py
#
# This source file is part of the asyncio-foundationdb open source project
#
# Copyright 2021 Amirouche Boubekki <amirouche@hyper.dev>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from uuid import uuid4
from uuid import UUID
from collections import namedtuple

from hashlib import blake2b as hasher
from more_itertools import sliced

import found


class BStoreException(found.BaseFoundException):
    pass


BSTORE_SUFFIX_HASH = [b'\x01']
BSTORE_SUFFIX_BLOB = [b'\x02']

BStore = namedtuple('BStore', ('name', 'prefix_hash', 'prefix_blob',))


def make(name, prefix):
    prefix = list(prefix)
    out = BStore(name, tuple(prefix + BSTORE_SUFFIX_HASH), tuple(prefix + BSTORE_SUFFIX_BLOB))
    return out


async def get_or_create(tx, bstore, blob):
    hash = hasher(blob).digest()
    key = found.pack((bstore.prefix_hash, hash))
    maybe_uid = await found.get(tx, key)
    if maybe_uid is not None:
        return UUID(bytes=maybe_uid)
    # Otherwise create the hash entry and store the blob with a new uid
    # TODO: Use a counter and implement a garbage collector, and implement
    # bstore.delete
    uid = uuid4()
    found.set(tx, key, uid.bytes)
    for index, slice in enumerate(sliced(blob, found.MAX_SIZE_VALUE)):
        found.set(tx, found.pack((bstore.prefix_blob, uid, index)), bytes(slice))
    return uid


async def get(tx, bstore, uid):
    key = found.pack((bstore.prefix_blob, uid))
    out = b''
    async for _, value in found.query(tx, key, found.next_prefix(key)):
        out += value
    if out == b'':
        raise BStoreException('BLOB should be in database: uid={}'.format(uid))
    return out



#
# found/eavstore.py
#
# This source file is part of the asyncio-foundationdb open source project
#
# Copyright 2021 Amirouche Boubekki <amirouche@hyper.dev>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from collections import namedtuple
from uuid import uuid4

import found


EAVStore = namedtuple('EAVStore', ('name', 'prefix_data', 'prefix_index'))

EAVSTORE_SUFFIX_DATA = [b'\x01']
EAVSTORE_SUFFIX_INDEX = [b'\x02']


def make(name, prefix):
    out = EAVStore(
        name,
        tuple(list(prefix) + EAVSTORE_SUFFIX_DATA),
        tuple(list(prefix) + EAVSTORE_SUFFIX_INDEX),
    )
    return out


def create(tx, eavstore, dict, uid=None):
    uid = uuid4() if uid is None else uid
    for key, value in dict.items():
        key = found.pack((eavstore.prefix_data, uid, key))
        found.set(tx, key, found.pack((value,)))

    for key, value in dict.items():
        key = found.pack((eavstore.prefix_index, key, value, uid))
        found.set(tx, key, b'')

    return uid


async def get(tx, eavstore, uid):
    out = dict()
    key = found.pack((eavstore.prefix_data, uid))
    async for key, value in found.query(tx, key, found.next_prefix(key)):
        prefix, _, key = found.unpack(key)
        out[key] = found.unpack(value)[0]
    return out


def remove(tx, eavstore, uid):
    dict = get(tx, eavstore, uid)
    for key, value in dict.items():
        key = tuple((eavstore.prefix_index, key))
        found.clear(tx, key)
    key = found.pack((eavstore.prefix_data, uid))
    found.clear(tx, key, found.next_prefix(key))


def update(tx, eavstore, uid, dict):
    remove(tx, eavstore, uid)
    create(tx, eavstore, dict, uid)


async def query(tx, eavstore, key, value):
    key = found.pack((eavstore.prefix_index, key, value))
    async for key, _ in found.query(tx, key, found.next_prefix(key)):
        _, _, _, uid = found.unpack(key)
        yield uid



#
# This source file was part of the FoundationDB open source project
# it was forked to implement the Python asyncio bindings in found project.
# see https://github.com/amirouche/asyncio-foundationdb
#
# Copyright 2013-2018 Apple Inc. and the FoundationDB project authors
# Copyright 2018-2021 Amirouche Boubekki <amirouche@hyper.dev>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import asyncio
import atexit
import struct
import threading
from collections import namedtuple

from found._fdb import lib
from found._fdb import ffi


class BaseFoundException(Exception):
    pass


class FoundException(BaseFoundException):
    """Exception raised when FoundationDB returns an error"""

    def __init__(self, code):
        super().__init__(code)
        self.code = code

    def __repr__(self):
        description = ffi.string(lib.fdb_get_error(self.code)).decode("utf-8")
        return "<FoundException {} ({})>".format(description, self.code)


def next_prefix(key):
    """Compute the smallest bytes sequence that does not start with key"""
    key = key.rstrip(b"\xff")
    if len(key) == 0:
        raise ValueError("Key must contain at least one byte not equal to 0xFF.")
    return key[:-1] + bytes((key[-1] + 1,))


def _check(code):
    if code != 0:
        raise FoundException(code)


_network_thread = None
_network_thread_reentrant_lock = threading.RLock()
_loop = None


class NetworkThread(threading.Thread):
    def run(self):
        _check(lib.fdb_run_network())


def _init():
    """Must be called after setting the event loop"""
    try:
        global _loop
        global _network_thread
        # init should only succeed once; If _network_thread is not None,
        # someone already successfully called init
        if _network_thread is not None:
            # This raise FoundException, even if there is no call to FDB API
            raise FoundException(2000)

        _network_thread = NetworkThread()
        _network_thread.daemon = True
        lib.fdb_setup_network()
        _network_thread.start()
        _loop = asyncio.get_event_loop()
    except Exception:  # noqa
        # We assigned _network_thread but did not succeed in init,
        # clear it out so the next caller has chance
        _network_thread = None
        raise


@atexit.register
def _stop_network():
    if _network_thread:
        _check(lib.fdb_stop_network())
        _network_thread.join()


STREAMING_MODE_WANT_ALL = -2
STREAMING_MODE_ITERATOR = -1
STREAMING_MODE_EXACT = 0
STREAMING_MODE_SMALL = 1
STREAMING_MODE_MEDIUM = 2
STREAMING_MODE_LARGE = 3
STREAMING_MODE_SERIAL = 4


@ffi.callback("void(FDBFuture *, void *)")
def on_transaction_commit(fdb_future, aio_future):
    aio_future = ffi.from_handle(aio_future)
    error = lib.fdb_future_get_error(fdb_future)
    if error == 0:
        _loop.call_soon_threadsafe(aio_future.set_result, None)
        lib.fdb_future_destroy(fdb_future)
    else:
        _loop.call_soon_threadsafe(aio_future.set_exception, FoundException(error))
        lib.fdb_future_destroy(fdb_future)


@ffi.callback("void(FDBFuture *, void *)")
def on_transaction_get_range(fdb_future, aio_future):
    aio_future = ffi.from_handle(aio_future)
    kvs = ffi.new("FDBKeyValue **")
    count = ffi.new("int *")
    more = ffi.new("fdb_bool_t *")
    error = lib.fdb_future_get_keyvalue_array(fdb_future, kvs, count, more)
    if error == 0:
        out = list()
        # XXX: Because ffi.gc doens't work this time and because
        # downstream the code expect a real bytes object; for the time
        # being we do a copy of the whole range iteration

        # total count of buffers for this key-value array
        copy = kvs[0][0:count[0]]

        for kv in copy:
            # XXX: manual unpacking because cffi doesn't known about packing
            # https://bitbucket.org/cffi/cffi/issues/364/make-packing-configureable
            memory = ffi.buffer(ffi.addressof(kv), 24)
            key_ptr, key_length, value_ptr, value_length = struct.unpack(
                "=qiqi", memory
            )
            key = ffi.buffer(ffi.cast("char *", key_ptr), key_length)
            value = ffi.buffer(ffi.cast("char *", value_ptr), value_length)
            # XXX: make a copy again
            out.append((key[:], value[:]))

        _loop.call_soon_threadsafe(aio_future.set_result, (out, count[0], more[0]))
        # since we make copies of the fdb_future result we don't need
        # to keep it around
        lib.fdb_future_destroy(fdb_future)
    else:
        _loop.call_soon_threadsafe(aio_future.set_exception, FoundException(error))
        lib.fdb_future_destroy(fdb_future)


Transaction = namedtuple('Transaction', ('pointer', 'db', 'snapshot'))


def _make_transaction(db, snapshot=False):
    # TODO: maybe free this double pointer?
    out = ffi.new("FDBTransaction **")
    lib.fdb_database_create_transaction(db.pointer, out)
    out = ffi.gc(out[0], lib.fdb_transaction_destroy)
    out = Transaction(out, db, snapshot)
    return out


@ffi.callback("void(FDBFuture *, void *)")
def _on_transaction_get_read_version(fdb_future, aio_future):
    aio_future = ffi.from_handle(aio_future)
    pointer = ffi.new("int64_t *")
    error = lib.fdb_future_get_int64(fdb_future, pointer)
    if error == 0:
        _loop.call_soon_threadsafe(aio_future.set_result, pointer[0])
    else:
        _loop.call_soon_threadsafe(aio_future.set_exception, FoundException(error))
    lib.fdb_future_destroy(fdb_future)


async def read_version(tx):
    fdb_future = lib.fdb_transaction_get_read_version(tx.pointer)
    aio_future = _loop.create_future()
    handle = ffi.new_handle(aio_future)
    lib.fdb_future_set_callback(fdb_future, _on_transaction_get_read_version, handle)
    out = await aio_future
    return out


@ffi.callback("void(FDBFuture *, void *)")
def _on_transaction_get(fdb_future, aio_future):
    aio_future = ffi.from_handle(aio_future)
    present = ffi.new("fdb_bool_t *")
    value = ffi.new("uint8_t **")
    value_length = ffi.new("int *")
    error = lib.fdb_future_get_value(fdb_future, present, value, value_length)
    if error == 0:
        if present[0] == 0:
            _loop.call_soon_threadsafe(aio_future.set_result, None)
            lib.fdb_future_destroy(fdb_future)
        else:
            # XXX: https://bitbucket.org/cffi/cffi/issues/380/ffibuffer-position-returns-a-buffer
            out = bytes(ffi.buffer(value[0], value_length[0]))
            _loop.call_soon_threadsafe(aio_future.set_result, out)
            lib.fdb_future_destroy(fdb_future)
    else:
        _loop.call_soon_threadsafe(aio_future.set_exception, FoundException(error))
        lib.fdb_future_destroy(fdb_future)


async def get(tx, key):
    assert isinstance(tx, Transaction)
    assert isinstance(key, bytes)
    fdb_future = lib.fdb_transaction_get(
        tx.pointer, key, len(key), tx.snapshot
    )
    aio_future = _loop.create_future()
    handle = ffi.new_handle(aio_future)
    lib.fdb_future_set_callback(fdb_future, _on_transaction_get, handle)
    out = await aio_future
    return out


KeySelector = namedtuple('KeySelector', ('key', 'or_equal', 'offset'))


def lt(key, offset=0):
    assert isinstance(key, bytes)
    return KeySelector(key, False, offset)


def lte(key, offset=0):
    assert isinstance(key, bytes)
    return KeySelector(key, True, offset)


def gt(key, offset=1):
    assert isinstance(key, bytes)
    return KeySelector(key, True, offset)


def gte(key, offset=1):
    assert isinstance(key, bytes)
    return KeySelector(key, False, 1)


async def query(tx, key, other, *, limit=0, mode=STREAMING_MODE_ITERATOR):
    key = key if isinstance(key, KeySelector) else gte(key)
    other = other if isinstance(other, KeySelector) else gte(other)
    if key.key < other.key:
        begin = key
        end = other
        reverse = False
    else:
        begin = other
        end = key
        reverse = True

    # the first read was fired off when the FDBRange was initialized
    iteration = 1
    snapshot = tx.snapshot
    while True:
        fdb_future = lib.fdb_transaction_get_range(
            tx.pointer,
            begin.key,
            len(begin.key),
            begin.or_equal,
            begin.offset,
            end.key,
            len(end.key),
            end.or_equal,
            end.offset,
            limit,
            0,
            mode,
            iteration,
            snapshot,
            reverse,
        )
        aio_future = _loop.create_future()
        handle = ffi.new_handle(aio_future)
        lib.fdb_future_set_callback(fdb_future, on_transaction_get_range, handle)
        kvs, count, more = await aio_future

        index = 0
        if count == 0:
            return

        for kv in kvs:
            yield kv

            index += 1
            if index == count:
                if not more or limit == count:
                    return

                iteration += 1
                if limit > 0:
                    limit -= count
                if reverse:
                    end = gte(kv[0])
                else:
                    begin = gt(kv[0])
        # loop!


@ffi.callback("void(FDBFuture *, void *)")
def _estimated_size_bytes_callback(fdb_future, aio_future):
    aio_future = ffi.from_handle(aio_future)
    pointer = ffi.new("int64_t *")
    error = lib.fdb_future_get_int64(fdb_future, pointer)
    if error == 0:
        _loop.call_soon_threadsafe(aio_future.set_result, pointer[0])
        lib.fdb_future_destroy(fdb_future)
    else:
        _loop.call_soon_threadsafe(aio_future.set_exception, FoundException(error))
        lib.fdb_future_destroy(fdb_future)


async def estimated_size_bytes(tx, begin, end):
    fdb_future = lib.fdb_transaction_get_estimated_range_size_bytes(
        tx.pointer,
        begin,
        len(begin),
        end,
        len(end)
    )
    aio_future = _loop.create_future()
    handle = ffi.new_handle(aio_future)
    lib.fdb_future_set_callback(fdb_future, _estimated_size_bytes_callback, handle)
    size = await aio_future
    return size


def set_read_version(tx, version):
    assert not tx.snapshot
    lib.fdb_transaction_set_read_version(tx.pointer, version)


def set(tx, key, value):
    assert isinstance(tx, Transaction)
    assert isinstance(key, bytes)
    assert isinstance(value, bytes)
    assert not tx.snapshot

    lib.fdb_transaction_set(tx.pointer, key, len(key), value, len(value))


def clear(tx, key, other=None):
    assert isinstance(tx, Transaction)
    assert isinstance(key, bytes)
    if other is None:
        lib.fdb_transaction_clear(tx.pointer, key, len(key))
    else:
        assert isinstance(other, bytes)
        lib.fdb_transaction_clear_range(tx.pointer, key, len(key), other, len(other))


async def _commit(tx):
    fdb_future = lib.fdb_transaction_commit(tx.pointer)
    aio_future = _loop.create_future()
    handle = ffi.new_handle(aio_future)
    lib.fdb_future_set_callback(fdb_future, on_transaction_commit, handle)
    await aio_future


MUTATION_ADD = 2
MUTATION_BIT_AND = 6
MUTATION_BIT_OR = 7
MUTATION_BIT_XOR = 8
MUTATION_MAX = 12
MUTATION_MIN = 13
MUTATION_SET_VERSIONSTAMPED_KEY = 14
MUTATION_SET_VERSIONSTAMPED_VALUE = 15
MUTATION_BYTE_MIN = 16
MUTATION_BYTE_MAX = 17


def _atomic(tx, opcode, key, param):
    lib.fdb_transaction_atomic_op(tx.pointer, key, len(key), param, len(param), opcode)


def add(tx, key, param):
    _atomic(tx, MUTATION_ADD, key, param)


def bit_and(tx, key, param):
    _atomic(tx, MUTATION_BIT_AND, key, param)


def bit_or(tx, key, param):
    _atomic(tx, MUTATION_BIT_OR, key, param)


def bit_xor(tx, key, param):
    _atomic(tx, MUTATION_BIT_XOR, key, param)


def max(tx, key, param):
    _atomic(tx, MUTATION_MAX, key, param)


def byte_max(tx, key, param):
    _atomic(tx, MUTATION_BYTE_MAX, key, param)


def min(tx, key, param):
    _atomic(tx, MUTATION_MIN, key, param)


def byte_min(tx, key, param):
    _atomic(tx, MUTATION_BYTE_MIN, key, param)


def set_versionstamped_key(tx, key, param):
    _atomic(tx, MUTATION_SET_VERSIONSTAMPED_KEY, key, param)


def set_versionstamped_value(tx, key, param):
    _atomic(tx, MUTATION_SET_VERSIONSTAMPED_VALUE, key, param)


@ffi.callback("void(FDBFuture *, void *)")
def _on_error_callback(fdb_future, aio_future):
    aio_future = ffi.from_handle(aio_future)
    error = lib.fdb_future_get_error(fdb_future)
    if error == 0:
        _loop.call_soon_threadsafe(aio_future.set_result, None)
    else:
        _loop.call_soon_threadsafe(aio_future.set_exception, FoundException(error))
    lib.fdb_future_destroy(fdb_future)


async def transactional(db, func, *args, snapshot=False, **kwargs):
    tx = _make_transaction(db, snapshot)
    while True:
        try:
            out = await func(tx, *args, **kwargs)
            await _commit(tx)
        except FoundException as exc:
            fdb_future = lib.fdb_transaction_on_error(tx.pointer, exc.code)
            aio_future = _loop.create_future()
            handle = ffi.new_handle(aio_future)
            lib.fdb_future_set_callback(fdb_future, _on_error_callback, handle)
            await aio_future  # may raise an exception
        else:
            return out


Database = namedtuple('Database', ('pointer',))


async def open(cluster_file=None):

    with _network_thread_reentrant_lock:
        if _network_thread is None:
            _init()

    out = ffi.new("FDBDatabase **")
    cluster_file = cluster_file if cluster_file is not None else ffi.NULL
    lib.fdb_create_database(cluster_file, out)
    out = ffi.gc(out[0], lib.fdb_database_destroy)
    out = Database(out)

    return out



# Copyright (C) 2020,2022 Amirouche A. Boubekki
#
# https://git.sr.ht/~amirouche/copernic
#
from uuid import uuid4

import fdb
import fdb.tuple

import found
from found.base import FoundException

from immutables import Map
import nstore



class VNStoreBase:
    pass


class VNStoreExcpetion(Exception):
    pass


class VNStore(VNStoreBase):

    def __init__(self, subspace, items):
        assert isinstance(subspace, (tuple, list))
        assert isinstance(items, (tuple, list))
        self._subspace = subspace
        self._items = items = list(items)
        # A change can have two key:
        #
        # - "message": that is a small description of the change
        #
        # - "significance": once the change is applied, it has an
        #    history significance VersionStamp that allows to order
        #    the changes.
        #
        self._changes = nstore.open(
            subspace + ['changes'],
            ('uid', 'key', 'value')
        )
        # self.tuples contains the tuples associated with the change
        # identifier (changeid) that created or removed the tuple
        # (alive?).
        self._tuples = nstore.open(
            subspace + ['tuples'],
            items + ['alive?', 'changeid']
        )

    def change_create(self, tr):
        # TODO: XXX: In theory, uuid4 can clash, replace with
        # VersionStamp.
        tr._vnstore_changeid = changeid = uuid4()
        # With significance as `None` the change is invisible to
        # VNStore.ask.
        self._changes.add(tr, changeid, 'significance', None)
        self._changes.add(tr, changeid, 'message', None)

        return changeid

    def change_continue(self, tr, changeid):
        tr._vnstore_changeid = changeid

    def change_message(self, tr, changeid, message):
        # Remove existing message if any
        bindings = self._changes.FROM(tr, changeid, 'message', nstore.var('message'))
        for binding in bindings:
            self._changes.delete(tr, changeid, 'message', binding['message'])
        # add message
        self._changes.add(tr, changeid, 'message', message)

    def change_apply(self, tr, changeid):
        # apply change by settings a verionstamp
        self._changes.delete(tr, changeid, 'significance', None)
        self._changes.add(tr, changeid, 'significance', fdb.tuple.Versionstamp())

    def ask(self, tr, *items):
        assert len(items) == len(self._items), "Incorrect count of ITEMS"
        # Complexity is O(n), where n is the number of times the exact
        # same ITEMS was added and deleted.  In pratice, n=0, n=1 or
        # n=2, and of course it always possible that it is more...
        bindings = self._tuples.FROM(tr, *items, nstore.var('alive?'), nstore.var('changeid'))
        found = False
        significance_max = fdb.tuple.Versionstamp(b'\x00' * 10)
        for binding in bindings:
            changeid = binding['changeid']
            significance = self._changes.FROM(
                tr,
                changeid, 'significance', nstore.var('significance')
            )
            significance = next(significance)
            significance = significance['significance']
            if (significance is not None) and (significance > significance_max):
                found = binding['alive?']
        return found

    def add(self, tr, *items):
        assert len(items) == len(self._items)
        if self.ask(tr, *items):
            # ITEMS already exists.
            return False
        # Add it
        items = list(items) + [True, tr._vnstore_changeid]
        self._tuples.add(tr, *items)
        return True

    def delete(self, tr, *items):
        assert len(items) == len(self._items)
        if not self.ask(tr, *items):
            # ITEMS does not exists.
            return False
        # Delete it
        items = list(items) + [False, tr._vnstore_changeid]
        self._tuples.add(tr, *items)
        return True

    def FROM(self, tr, *pattern, seed=Map()):  # seed is immutable
        """Yields bindings that match PATTERN"""
        assert len(pattern) == len(self._items), "invalid item count"
        # TODO: validate that pattern does not have variables named
        # `alive?` or `changeid`.

        def bind(pattern, binding):
            for item in pattern:
                if isinstance(item, Variable):
                    yield binding[item.name]
                else:
                    yield item

        # The complexity really depends on the pattern.  A pattern
        # only made of variables will scan the whole database.  In
        # practice, the user will seldom do time traveling queries, so
        # it should rarely hit this code path.
        pattern = list(pattern) + [nstore.var('alive?'), nstore.var('changeid')]
        bindings = self._tuples.FROM(tr, *pattern, seed=seed)
        for binding in bindings:
            if not binding['alive?']:
                # The associated tuple is dead, so the bindings are
                # not valid in all cases.
                continue
            elif self.ask(self, *bind(pattern, binding)):
                # The bound pattern exist, so the bindings are valid
                binding = binding.delete('alive?')
                binding = binding.delete('changeid')
                yield binding
            else:
                continue

    def where(self, tr, *pattern):
        assert len(pattern) == len(self._items), "invalid item count"

        def _where(iterator):
            for bindings in iterator:
                # bind PATTERN against BINDINGS
                bound = []
                for item in pattern:
                    # if ITEM is variable try to bind
                    if isinstance(item, Variable):
                        try:
                            value = bindings[item.name]
                        except KeyError:
                            # no bindings
                            bound.append(item)
                        else:
                            # pick the value from bindings
                            bound.append(value)
                    else:
                        # otherwise keep item as is
                        bound.append(item)
                # hey!
                yield from self.FROM(tr, *bound, seed=bindings)

        return _where



open = VNStore


def select(seed, *wheres):
    out = seed
    for where in wheres:
        out = where(out)
    return out



#
# found/pstore.py
#
# This source file is part of the asyncio-foundationdb open source project
#
# Copyright 2021 Amirouche Boubekki <amirouche@hyper.dev>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import random
import asyncio
from operator import itemgetter
from uuid import uuid4
from collections import namedtuple
from collections import Counter

import found
from found import nstore

import zstandard as zstd


class PStoreException(found.BaseFoundException):
    pass


try:
    FOUND_PSTORE_SAMPLE_COUNT = int(os.environ['FOUND_PSTORE_SAMPLE_COUNT'])
except (KeyError, ValueError):
    FOUND_PSTORE_SAMPLE_COUNT = 1337


PSTORE_SUFFIX_TOKENS = [b'\x01']
PSTORE_SUFFIX_INDEX = [b'\x02']
PSTORE_SUFFIX_COUNTERS = [b'\x03']

PStore = namedtuple('PStore', ('name', 'tokens', 'prefix_index', 'prefix_counters', 'pool'))


def make(name, prefix):
    prefix = list(prefix)
    prefix_tokens = tuple(prefix + PSTORE_SUFFIX_TOKENS)
    tokens = nstore.make('{}/token'.format(name), prefix_tokens, 2)
    out = PStore(
        name,
        # Use nstore with n=2 to be able to go from a string token to an uid,
        # and back from an uid to a token string.
        tokens,
        # TODO: Replace with a multi-dict (mstore) dedicated store.
        # The value is always empty.
        tuple(prefix + PSTORE_SUFFIX_INDEX),
        # That will map bag uid to a counter serialized to json and
        # compressed with zstd. It is a good old key-value store.
        tuple(prefix + PSTORE_SUFFIX_COUNTERS),
        None,
    )
    return out


async def index(tx, store, docuid, counter):
    # translate keys that are string tokens, into uuid4 bytes with
    # store.tokens
    tokens = dict()
    for string, count in counter.items():
        query = nstore.select(tx, store.tokens, string, nstore.var('uid'))
        try:
            uid = await query.__anext__()
        except StopAsyncIteration:
            uid = uuid4()
            nstore.add(tx, store.tokens, string, uid)
        else:
            uid = uid['uid']
        tokens[uid] = count

    # store tokens to use later during search for filtering
    found.set(
        tx,
        found.pack((store.prefix_counters, docuid)),
        zstd.compress(found.pack(tuple(tokens.items())))
    )

    # store tokens keys for candidate selection
    for token in tokens:
        found.set(tx, found.pack((store.prefix_index, token, docuid)), b'')


async def _keywords_to_token(tx, tokens, keyword):
    query = nstore.select(tx, tokens, keyword, nstore.var('uid'))
    try:
        uid = await query.__anext__()
    except StopAsyncIteration:
        return None
    else:
        uid = uid['uid']
        return uid


async def _token_to_size(tx, prefix_index, token):
    key = found.pack((prefix_index, token))
    out = await found.estimated_size_bytes(tx, key, found.next_prefix(key))
    return out


async def _prepare(tx, prefix, candidates, keywords):
    for candidate in candidates:
        out = await found.get(tx, found.pack((prefix, candidate)))
        yield (candidate, keywords, out)


def _score(args):
    candidate, keywords, counter = args
    counter = dict(found.unpack(zstd.decompress(counter)))
    score = 0
    for keyword in keywords:
        try:
            count = counter[keyword]
        except KeyError:
            return None
        else:
            score += count
    return (candidate, score)


def _filter(hits):
    def wrapped(args):
        if args is not None:
            candidate, score = args
            hits[candidate] = score
    return wrapped


async def massage(tx, store, candidate, keywords, hits):
    score = 0
    counter = await found.get(tx, found.pack((store.prefix_counters, candidate)))
    # TODO: replace the dictionary and the following for loop with
    # a single iteration over the counter, using zigzag algorithm.
    counter = dict(found.unpack(zstd.decompress(counter)))
    for keyword in keywords:
        try:
            count = counter[keyword]
        except KeyError:
            return None
        else:
            score += count
    hits[candidate] = score


async def search(tx, store, keywords, limit=13):
    coroutines = (_keywords_to_token(tx, store.tokens, keyword) for keyword in keywords)
    keywords = await asyncio.gather(*coroutines)
    # If a keyword is not present in store.tokens, then there is no
    # document associated with it, hence there is no document that
    # match that keyword, hence no document that has all the requested
    # keywords. Return an empty counter.
    if any(keyword is None for keyword in keywords):
        return list()

    # Select seed token
    coroutines = (_token_to_size(tx, store.prefix_index, token) for token in keywords)
    sizes = await asyncio.gather(*coroutines)
    _, seed = min(zip(sizes, keywords), key=itemgetter(0))

    # Select candidates
    candidates = []
    key = found.pack((store.prefix_index, seed))
    query = found.query(tx, key, found.next_prefix(key))

    async for key, _ in query:
        _, _, uid = found.unpack(key)
        candidates.append(uid)

    # XXX: 500 was empirically discovered, to make it so that the
    #      search takes less than 1 second or so.
    if len(candidates) >= FOUND_PSTORE_SAMPLE_COUNT:
        candidates = random.sample(candidates, FOUND_PSTORE_SAMPLE_COUNT)

    # score, filter and construct hits aka. massage
    hits = Counter()

    coroutines = (massage(tx, store, c, keywords, hits) for c in candidates)
    await asyncio.gather(*coroutines)

    out = hits.most_common(limit)

    return out
